import ctypes

# must be greater than 0.0, less than 3000.0, the funnier the better
FUNNY_NUMBER = 420.69

# binary representations of the FUNNY_NUMBER according to RFC 8794, section 7.3.
FLOATS = {
    0: b'',
    4: bytes(ctypes.c_float(FUNNY_NUMBER))[::-1],
    8: bytes(ctypes.c_double(FUNNY_NUMBER))[::-1],
}

# Segment/Info/Duration path IDs of tags, converted from VINT format from here:
# https://github.com/ietf-wg-cellar/matroska-specification/blob/master/ebml_matroska.xml
ELEMENT_ID_PATH = [139690087, 88713574, 1161]

VINT_MAX_WIDTH = 4  # max amount of bytes on VINT_WIDTH and VINT_MARKER parts


def bytes2bitstr(data: bytes) -> str:
    nbytes = len(data)
    return format(int.from_bytes(data, "big"), f'0{nbytes * 8}b')


def read_vint(data: bytes) -> tuple[int, int]:
    """
    Reads Variable-Size Integer from the start of data
    according to RFC 8794 section 4.
    """
    vint_length = bytes2bitstr(data[:VINT_MAX_WIDTH]).index("1") + 1
    vint_value = int(bytes2bitstr(data[:vint_length])[vint_length:], 2)
    return vint_length, vint_value


def enter_element(s: int, f: int, data: bytes) -> tuple[int, int]:
    vint_len, element_len = read_vint(data[s:f])
    s += vint_len
    f = s + element_len
    return s, f


def skip_element(s: int, f: int, data: bytes) -> tuple[int, int]:
    vint_len, element_len = read_vint(data[s:f])
    s += vint_len + element_len
    return s, f


def find_duration_tag(data: bytes) -> tuple[int, int]:
    # start and finish of the current working byte sequence in data
    s: int = 0
    f: int = len(data)
    i: int = 0

    while s != f:
        vint_len, element_id = read_vint(data[s:f])
        s += vint_len

        if element_id == ELEMENT_ID_PATH[i]:
            s, f = enter_element(s, f, data)
            i += 1
            if i == len(ELEMENT_ID_PATH):
                return s, f
        else:
            s, f = skip_element(s, f, data)
    else:
        print("Duration value was not found")
        exit(1)


def spoof_duration(data: bytes) -> bytearray:
    s, f = find_duration_tag(data)
    data = bytearray(data)
    data[s:f] = FLOATS[f - s]
    return data


def spoof_file_duration(source_file, result_file):
    print(f"Spoofing duration of {source_file}...")

    with open(source_file, 'rb') as source:
        data = source.read()

    with open(result_file, 'wb') as result:
        result.write(spoof_duration(data))

    print("Done.")



import copy
from pathlib import Path
import subprocess

import pydantic

from .config_model import CmdConfig, CmdParams, ArgsPlaceholder
from . import spoofer

SIZE_256KB = 1 << 18


class ArgsQueue:

    def __init__(self, argv):
        self.args: list[str] = argv[::-1]

    def pop_next(self) -> str | None:
        try:
            return self.args.pop()
        except IndexError:
            return None

    def pop_next_n(self, n: int) -> list[str] | None:
        if len(self.args) < n:
            return None
        return [self.args.pop() for _ in range(n)]


class RunInfo(pydantic.BaseModel):
    input: str
    output: str
    guess_value: str
    guess_iterations: str
    guess_min: str | None
    guess_max: str | None


def generate_tmp_file_path(output: str, suffix: str = ""):
    output_path = Path(output)
    tmp_dir = output_path.parent / "tmp"
    tmp_dir.mkdir(exist_ok=True)
    return (tmp_dir / output_path.name).with_suffix(f".i{suffix}.webm")


def run_ffmpeg_commands(config: CmdConfig, params: CmdParams, output: Path):
    for command in map(list.copy, config.passes):
        ph = command.index(ArgsPlaceholder())
        command[ph:ph + 1] = params.args
        params.placeholders["output"] = str(output)

        command = [arg.format(**params.placeholders) for arg in command]

        print(f"$", *command)
        if returncode := subprocess.run(command).returncode:
            exit(returncode)


def parse_command_args(config: CmdConfig,
                       argv: list[str]) -> dict[str, list[str]]:
    args = ArgsQueue(argv)
    alias_dict = config.map_flag_aliases()

    flag_args_dict: dict[str, list[str]] = {}

    while flag_alias := args.pop_next():
        flag = alias_dict.get(flag_alias)
        if not flag:
            raise ValueError(f"Unknown flag: '{flag_alias}'")

        if (flag_args := args.pop_next_n(flag.nargs)) is None:
            raise ValueError(f"Not enough arguments for a '{flag_alias}'")

        if flag.name in flag_args_dict:
            raise ValueError(f"Duplicate flag: '{flag_alias}'")

        flag_args_dict[flag.name] = flag_args

    return flag_args_dict


def guess_value(config: CmdConfig, cmd_params: CmdParams, run_info: RunInfo):
    guess_value_flag = config.values.get(run_info.guess_value)

    if guess_value_flag is None:
        raise ValueError(f"Unknown guess value: '{run_info.guess_value}'")

    if guess_value_flag.guess_params is None:
        raise ValueError(f"Value '{run_info.guess_value}' is not guessable")

    params = guess_value_flag.guess_params

    gtype = params.type.get_constructor()
    gmin = gtype(run_info.guess_min or params.min)
    gmax = gtype(run_info.guess_max or params.max)

    try:
        g_it = int(run_info.guess_iterations)
    except ValueError:
        raise ValueError(f"Guess iterations is '{run_info.guess_iterations}',"
                         " should be an integer")

    best_valid_option = None

    print(f"Starting to guess {run_info.guess_value} on interval"
          f" [{gmin}, {gmax}] in {g_it} iterations...")

    for i in range(g_it):
        mean_value = gtype((gmin + gmax) / 2)
        fvalue = params.type.format(mean_value)
        cmd_params_i = copy.deepcopy(cmd_params)
        guess_value_flag.parse(run_info.guess_value, [fvalue], cmd_params_i)

        path = generate_tmp_file_path(run_info.output, f"_{fvalue}")

        print(f"\nIteration №{i + 1}, running ffmpeg"
              f" with {run_info.guess_value} = {fvalue}:")
        run_ffmpeg_commands(config, cmd_params_i, output=path)
        result_size = path.stat().st_size

        if result_size == SIZE_256KB:
            print(f"Result is ideal size.\n")
            best_valid_option = path
            break
        elif result_size < SIZE_256KB:
            print(f"Result is smaller than limit, increasing lower limit...\n")
            best_valid_option = path
            cut_from = "below"
        else:
            print(f"Result is bigger than limit, decreasing upper limit...\n")
            cut_from = "above"

        if (cut_from == "below") ^ params.scaling.is_inverse():
            gmin = mean_value
        else:
            gmax = mean_value

    if best_valid_option is None:
        print("Script did not find settings for a sufficiently small final"
              " file, try increasing the number of iterations"
              " or reducing the quality settings.")
        exit(1)

    return best_valid_option


def convert_video(config: CmdConfig, argv: list[str]):
    flag_args_dict = parse_command_args(config, argv)
    cmd_params = CmdParams(config)

    for flag_name, flag in config.flag_dict.items():
        flag.parse(flag_name, flag_args_dict.get(flag_name), cmd_params)

    cmd_params.set_default_output()
    run_info = RunInfo(**cmd_params.placeholders)

    if run_info.guess_value == "none":
        output_file = generate_tmp_file_path(run_info.output)
        run_ffmpeg_commands(config, cmd_params, output=output_file)
    else:
        output_file = guess_value(config, cmd_params, run_info)

    efficiency = output_file.stat().st_size / SIZE_256KB
    print(f"Final file uses {efficiency:.2%}"
          " of avaliable sticker file size.")

    spoofer.spoof_file_duration(output_file, run_info.output)


from collections import Counter, ChainMap
import itertools
from pathlib import Path
from typing import ClassVar, Iterable, TypeVar

import pydantic
import enum

T = TypeVar("T")


def get_duplicates_list(l: Iterable[T]) -> list[T]:
    return [item for item, count in Counter(l).items() if count > 1]


class CmdParams:

    def __init__(self, config: "CmdConfig") -> None:
        self.args: list[str] = config.required_args
        self.placeholders: dict[str, str] = {}

    def set_default_output(self):
        if "output" not in self.placeholders:
            output = Path(self.placeholders["input"]).with_suffix(".webm")
            print(f"Output file defaults to '{output}'.")
            self.placeholders["output"] = str(output)


class CmdFlag(pydantic.BaseModel):
    name: str = "..."  # post-initialized in the CmdConfig._add_flag_names
    description: str
    aliases: list[str]
    nargs: ClassVar[int]

    class Config:
        extra = "forbid"

    def get_flag_alias_dict(self) -> dict[str, "CmdFlag"]:
        return dict.fromkeys(self.aliases, self)

    def parse(self, flag_name: str, flag_args: list[str] | None,
              cmd_params: CmdParams) -> None:
        raise NotImplementedError

    def _print_flags_and_description(self):
        flags = ", ".join(self.aliases)
        print(f"{flags:<23} {self.description}")

    def print_help(self):
        self._print_flags_and_description()


class ArgsPlaceholder(pydantic.BaseModel):

    def format(self, **_kwargs):
        raise NotImplementedError

    class Config:
        extra = "forbid"


class EnumOption(pydantic.BaseModel):
    description: str
    default: bool = False
    args: list[str] = []
    placeholders: dict[str, str] = {}

    def apply(self, cmd_params: CmdParams):
        cmd_params.args += self.args
        cmd_params.placeholders |= self.placeholders

    def print_help(self, name):
        default_marker = "(default)" if self.default else ""
        print(f"    {name:<23} {self.description} {default_marker}")


class EnumFlag(CmdFlag):
    options: dict[str, EnumOption]
    nargs: ClassVar[int] = 1

    @pydantic.root_validator
    def only_one_default_option(cls, values):
        options: dict[str, EnumOption] = values.get("options")
        defaults = [name for name, option in options.items() if option.default]
        if not defaults:
            raise ValueError(f"There is no defult options"
                             f" in '{values.get('name')}' flag, should be 1")

        if (n := len(defaults)) > 1:
            raise ValueError(f"There is {n} default options"
                             f" in '{values.get('name')}' flag:"
                             f" {', '.join(defaults)}, should be only 1")
        return values

    @property
    def default_option(self) -> EnumOption:
        for option in self.options.values():
            if option.default:
                return option
        raise  # default option is always present

    def parse(self, flag_name, flag_args, cmd_params):
        if not flag_args:
            self.default_option.apply(cmd_params)
            return

        option_name = flag_args[0]
        option = self.options.get(option_name)
        if not option:
            raise ValueError(f"Unknown '{flag_name}' option: '{option_name}'")

        option.apply(cmd_params)

    def print_help(self):
        print()
        self._print_flags_and_description()
        for option_name, option in self.options.items():
            option.print_help(option_name)


class SwitchFlag(CmdFlag):
    inverted: bool = False
    args: list[str]
    nargs: ClassVar[int] = 0

    def parse(self, _flag_name, flag_args, cmd_params):
        if (flag_args == []) ^ self.inverted:
            cmd_params.args += self.args


class ValueType(enum.Enum):
    INT = "int"
    FLOAT = "float"

    def get_constructor(self):
        if self == self.INT:
            return int
        return float

    def format(self, n: int | float):
        if self == self.INT:
            return f"{int(n):d}"
        return f"{n:.2f}"


class Scaling(enum.Enum):
    DIRECT = "direct"
    INVERSE = "inverse"

    def is_inverse(self):
        return self == self.INVERSE


class GuessParams(pydantic.BaseModel):
    min: int | float
    max: int | float
    type: ValueType
    scaling: Scaling


class ValueFlag(CmdFlag):
    args: list[str] = []
    optional: bool = False
    default_value: str | None
    guess_params: GuessParams | None = None
    nargs: ClassVar[int] = 1

    def parse(self, flag_name, flag_args, cmd_params):
        if not flag_args:
            if not self.optional:
                raise ValueError(f"Required value '{flag_name}'"
                                 " is not present")
            if self.default_value is not None:
                cmd_params.args += self.args
                cmd_params.placeholders[flag_name] = self.default_value
            return

        cmd_params.args += self.args
        cmd_params.placeholders[flag_name] = flag_args[0]

    def print_help(self):
        flags = ", ".join(self.aliases)
        guessable_marker = "(guessable)" if self.guess_params else ""
        print(f"{flags:<23} {self.description} {guessable_marker}")


class CmdConfig(pydantic.BaseModel):
    passes: list[list[str | ArgsPlaceholder]]
    required_args: list[str]
    enums: dict[str, EnumFlag] = {}
    switches: dict[str, SwitchFlag] = {}
    values: dict[str, ValueFlag] = {}

    @pydantic.root_validator
    def _add_flag_names(cls, values: dict[str, dict[str, CmdFlag]]):
        for flag_dict_name in ("enums", "switches", "values"):
            for name, cmdflag in values[flag_dict_name].items():
                cmdflag.name = name
        return values

    @property
    def flag_dicts(self) -> tuple[dict[str, CmdFlag], ...]:
        return (self.enums, self.switches, self.values)  # type: ignore

    @property
    def flag_dict(self) -> dict[str, CmdFlag]:
        flag_name_list = itertools.chain(*map(dict.keys, self.flag_dicts))

        if duplicates := get_duplicates_list(flag_name_list):
            raise ValueError(f"Duplicate flag names: {', '.join(duplicates)}.")

        return dict(ChainMap(*self.flag_dicts))

    def map_flag_aliases(self) -> dict[str, CmdFlag]:
        alias_dict = {}

        for flag_name, flag in self.flag_dict.items():
            for alias in flag.aliases:
                if alias in alias_dict:
                    raise ValueError(f"Duplicate alias '{alias}'"
                                     f" for the {flag_name} flag")

            alias_dict.update(flag.get_flag_alias_dict())

        return alias_dict

    def print_help(self):
        gvalues_list = [i.name for i in self.values.values() if i.guess_params]
        gvalues_list.append("none")
        guessable_values = "\n\nGuessable values: " + ", ".join(gvalues_list)

        section_headers = (
            "Enum flags. Usage: '--flag <OPTION>'",
            "Switch flags. Usage: '--flag' to turn on",
            "Value flags. Usage: '--flag <VALUE>'" + guessable_values,
        )

        for section, header in zip(self.flag_dicts, section_headers):
            print(f"\n{header}\n")
            for flag in section.values():
                flag.print_help()
            print()



import os
from pathlib import Path
import sys
import shutil

import pydantic
import tomli

from .converter import CmdConfig, convert_video
from .spoofer import spoof_file_duration

DEFAULT_CONFIG_PATH = Path(__file__).parent / "default_config.toml"
CONFIG_PATH = Path(__file__).parent / "config.toml"

TGRADISH_CMD = "tgradish"


def check_file_existence(file_path: str):
    if not Path(file_path).exists():
        print(f"File '{file_path}' is not exists.")
        exit(1)


def set_default_config():
    print("Setting default config...")
    shutil.copyfile(DEFAULT_CONFIG_PATH, CONFIG_PATH)
    print("Done.\n")


def get_config() -> CmdConfig:
    if not Path(CONFIG_PATH).exists():
        set_default_config()
    return CmdConfig(**tomli.load(open(CONFIG_PATH, "rb")))


def print_help():
    print(
        f"Usage: {TGRADISH_CMD} [ACTION] [ARGUMENTS]\n",
        "-h, --help          shows this message",
        "[ACTION] --help     shows help for specific action",
        "\nActions:\n",
        "convert   converts source file to videosticker with spoofed duration",
        "spoof     spoofs .webm file duratiob",
        "config    manages tgradish config file",
        sep="\n",
    )


def convert_cmd(args: list[str]):
    config = get_config()

    if not args or args[0] in ("-h", "--help"):
        print(TGRADISH_CMD, "convert [FLAGS]\n")
        config.print_help()
        exit()

    try:
        convert_video(config, args)
    except (ValueError, pydantic.error_wrappers.ValidationError) as err:
        print(f"Converter error:\n{err}")
        exit(1)


def spoof_cmd(args: list[str]):
    help_str = TGRADISH_CMD + " spoof [INPUT FILE] [OUTPUT FILE]"

    if not args or args[0] in ("-h", "--help"):
        print(help_str, "\n\nThat's it.")
        exit()
    elif len(args) != 2:
        print("Incorrect arguments format.\n\n", help_str, sep="")
        exit(1)

    input_file, output_file = args
    check_file_existence(input_file)
    spoof_file_duration(input_file, output_file)


def config_cmd(args: list[str]):
    help_str = TGRADISH_CMD + " config [ACTION]"

    if not args or args[0] in ("-h", "--help"):
        print(
            help_str,
            "\nActions:\n",
            "copyfrom [FILE] copies config from file",
            "restore         restores default config",
            "showpath        shows path to config",
            sep="\n",
        )

    elif args[0] == "copyfrom":
        if len(args) != 2:
            print(f"Usage: {TGRADISH_CMD} config copyfrom [FILE]")
            exit(1)
        print(f"Copying config file from {args[1]}...")
        shutil.copyfile(args[1], CONFIG_PATH)
        print("Done.")
        exit()

    elif args[0] == "restore":
        set_default_config()

    elif args[0] == "showpath":
        print(Path(CONFIG_PATH).resolve())

    else:
        print("Invalid config action.")
        exit(1)


def main():
    match sys.argv[1:]:
        case [] | ["-h" | "--help", *_]:
            print_help()
        case ["convert", *args]:
            convert_cmd(args)
        case ["spoof", *args]:
            spoof_cmd(args)
        case ["config", *args]:
            config_cmd(args)
